# Import required libraries
import numpy as np
import pandas as pd
import math
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
from sklearn.metrics import roc_auc_score, roc_curve
from ML_medic_kit.model_selection import create_classifier
import shap
import seaborn as sns
sns.set(color_codes=True)


def plot_class_distribution(data, column, title='Class Distribution'):
    """
    Plot the count of each value in the specified column.

    Parameters:
        data (DataFrame): The DataFrame containing the data.
        column (str): The column name for which the distribution
                      is to be plotted.
        title (str): The title of the plot.
        palette (str):  The color palette to use for the plot.

    Returns:
        None: Shows a count plot.
    """
    plt.figure(figsize=(8, 5))
    sns.countplot(x=data[column])

    # Add title and labels
    plt.title(title)
    plt.xlabel(column)
    plt.ylabel('Count')

    # Show the plot
    plt.show()


def plot_explained_variance(X_scaled):
    """
    Plots the explained variance and cumulative explained variance
    for different numbers of components.

    Parameters:
        X_scaled (array): The scaled feature data.
    """
    pca = PCA().fit(X_scaled)
    explained_variance = pca.explained_variance_ratio_
    cumulative_explained_variance = explained_variance.cumsum()

    plt.figure(figsize=(14, 5))

    # Subplot 1: Explained Variance
    plt.subplot(1, 2, 1)
    plt.bar(range(1, len(explained_variance) + 1),
            explained_variance, alpha=0.5, align='center',
            label='Individual explained variance')
    plt.ylabel('Explained variance ratio')
    plt.xlabel('Principal component index')
    plt.title('Explained Variance by Different Principal Components')

    # Subplot 2: Cumulative Explained Variance
    plt.subplot(1, 2, 2)
    plt.step(range(1, len(cumulative_explained_variance) + 1),
             cumulative_explained_variance, where='mid',
             label='Cumulative explained variance')
    plt.ylabel('Cumulative explained variance ratio')
    plt.xlabel('Number of components')
    plt.title('Cumulative Explained Variance by Components')

    plt.show()


def correlation_heatmap(data, cmap='coolwarm', filt=['_se', '_worst'],
                        save_fig=False, filepath='n/a'):
    """
    Plot correlation heatmap of the data.

    Parameters:
        data (Dataframe): Dataset to plot.
        cmap (str): optional (default='coolwarm')
            Colour map of figure.
        filt (list): optional (default=['_se', '_worst'])
            Filters out columns with certain names not to plot.
        save_fig (bool): optional (default=False)
            If figure should be saved to specified location.
        filepath (str): optional (default='n/a')
            Filepath to save figure to.

    Returns:
        None: Plots a correlation heatmap.
    """
    # Create a copy of the data to avoid modifying original
    data_plot = data.copy()

    # Filters out columns with certain names
    for drop in filt:
        data_plot.drop(list(data_plot.filter(regex=drop)),
                       axis=1, inplace=True)

    # Calculates correlation
    correlation_matrix = data_plot.corr()

    # Sets figure with specified figure size
    plt.figure(figsize=(8, 6))
    # Plots heatmap
    mask = np.triu(np.ones_like(correlation_matrix, dtype=bool))
    sns.heatmap(correlation_matrix, annot=True, cmap=cmap,
                fmt='.2f', linewidths=0.5, mask=mask)

    plt.xticks(rotation=90)
    plt.title('Correlation Heatmap')

    # Saves figure if required and prevents axis
    # labels being cut off with tight layout
    if save_fig:
        plt.tight_layout()
        plt.savefig(filepath, dpi=300)

    plt.show()


def plot_roc(models, X_test, y_test, title_prefix='',
             save_fig=False, filepath='n/a'):
    """
    Plots ROC curves for given models.

    Parameters:
        models (dict): Dictionary with model objects.
        X_test: Test set.
        y_test: Actual outcomes for the test set.
        title_prefix (str): Prefix for the plot titles.
        save_fig (bool): optional (default=False)
            If figure should be saved to specified location.
        filepath (str): optional (default='n/a')
            Filepath to save figure to.

    Returns:
        None: ROC curve plot.
    """
    num_classifiers = len(models)
    fig, axes = plt.subplots(1, num_classifiers,
                             figsize=(6 * num_classifiers, 5))

    # If there is only one classifier, convert axes to a list
    if num_classifiers == 1:
        axes = [axes]

    for i, (classifier, model) in enumerate(models.items()):
        ax = axes[i]

        # Check if the model has the predict_proba method
        if not hasattr(model, 'predict_proba'):
            raise ValueError((f'The model {classifier} does not '
                              f'support probability predictions.'))

        # Calculate ROC curve and AUC
        y_pred_proba = model.predict_proba(X_test)[:, 1]
        fpr, tpr, _ = roc_curve(y_test, y_pred_proba)
        auc_score = round(roc_auc_score(y_test, y_pred_proba), 4)

        # Plot the ROC curve on it's subplot
        ax.plot(fpr, tpr, label=f'AUC={auc_score}')
        ax.plot([0, 1], [0, 1], 'k--')  # Diagonal line
        ax.set_xlabel('1 – Specificity (False Positive Rate)')
        ax.set_ylabel('Sensitivity (True Positive Rate)')
        ax.set_title(f'{title_prefix}ROC Curve for {classifier}')
        ax.legend()
        ax.grid(True)

    plt.tight_layout()

    # Saves figure if required
    if save_fig:
        plt.savefig(filepath, dpi=300)

    plt.show()


def variables_based_outcome(data, var1, var2, cmap='PiYG',
                            save_fig=False, filepath='n/a'):
    """
    Plots two different variables from the data against each
    other and colour-codes them based on the disease outcome.

    Parameters:
        data (Dataframe): Dataset to plot.
        var1 + var2 (str): Names of variables in data to plot.
        cmap (str): optional (default='PiYG')
            Colour map to use in plot.
        save_fig (bool): optional (default=False)
            If figure should be saved to specified location.
        filepath (str): optional (default='n/a')
            Filepath to save figure to.

    Returns:
        None: Plots scatterplot of 2 variables.
    """
    # Sets figure with specified figure size
    fig = plt.figure(figsize=(5, 5))
    # Plots variables and colour-codes based on outcome
    plt.scatter(data[var1], data[var2], c=data['outcome'],
                cmap=cmap, s=5)

    # Axis labels
    var1_label = var1.title().replace('_', ' ')
    var2_label = var2.title().replace('_', ' ')
    plt.xlabel(var1_label)
    plt.ylabel(var2_label)

    # Saves figure if required
    if save_fig:
        plt.savefig(filepath, dpi=300)

    plt.show()


def histogram_based_outcome(data, label=['M', 'B'], colours=['b', 'r'],
                            filt=['_se', '_worst'], figsize=(8, 10),
                            save_fig=False, filepath='n/a'):
    """
    Plots stacked histogram for all features in the data based on the
    disease outcome.

    Parameters:
        data (Dataframe): Dataset to plot.
        label (list of str): optional (default=['M', 'B'])
            Names of outcome to label.
        colours (list of str): optional (default=['b', 'r'])
            Colours that represent outcomes.
        filt (list): optional (default=['_se', '_worst'])
            Filters out columns with certain names not to plot.
        save_fig (bool): optional (default=False)
            If figure should be saved to specified location.
        filepath (str): optional (default='n/a')
            Filepath to save figure to.

    Returns:
        None: Plots histogram of variables.
    """
    # Create a copy of the data to avoid modifying original
    data_plot = data.copy()

    # Sets up outcome to label graph
    outcome_1 = data_plot[data_plot['outcome'] == 1]
    outcome_0 = data_plot[data_plot['outcome'] == 0]

    # Locates only the features to plot
    # and filters out columns with certain names
    for drop in filt:
        data_plot.drop(list(data_plot.filter(regex=drop)),
                       axis=1, inplace=True)

    # Finds all features excluding the outcome column
    features = list(data_plot.loc[:, data_plot.columns != 'outcome'])
    nrows = math.ceil(len(features)/2)
    # Prevents error if not enough features to fill both sides
    if len(features) % 2 == 0:
        odd_num = False
    else:
        odd_num = True

    fig, axes = plt.subplots(nrows=nrows, ncols=2, figsize=figsize)
    axes = axes.ravel()

    # Plots each feature
    for ind, ax in enumerate(axes):
        if odd_num and ind >= len(features):
            break
        feats = data_plot[features[ind]]
        binwidth = (max(feats) - min(feats)) / 60
        ax.hist([outcome_1[features[ind]], outcome_0[features[ind]]],
                bins=np.arange(min(feats), max(feats)+binwidth, binwidth),
                stacked=True, label=label, color=colours)
        ax.legend(loc='best')
        label_name = features[ind].title().replace('_', ' ')
        ax.set_xlabel(label_name)

    plt.tight_layout()

    # Saves figure if required
    if save_fig:
        plt.savefig(filepath, dpi=300)

    plt.show()


def plot_confusion_matrix(results, title_prefix='',
                          save_fig=False, filepath='n/a'):
    """
    Plots confusion matrices for the given results.

    Parameters:
        results (dict): The results dictionary containing confusion matrices.
        title_prefix (str): optional (default='')
            The prefix title of each plot.
        save_fig (bool): optional (default=False)
            If figure should be saved to specified location.
        filepath (str): optional (default='n/a')
            Filepath to save figure to.

    Returns:
        None: Plots confusion matrices for each model.
    """
    # If there is only one classifier, convert axes to a list
    num_classifiers = len(results)
    num_rows = 1
    num_cols = num_classifiers

    # Create subplots
    fig, axes = plt.subplots(num_rows, num_cols, figsize=(15, 5))

    if num_classifiers == 1:
        axes = [axes]

    # Iterate over classifiers and their metrics
    for i, (classifier, metrics) in enumerate(results.items()):

        # Extract the confusion matrix from metrics
        confusion_matrix = np.array(metrics['confusion_matrix'])

        ax = axes[i]

        # Plot the confusion matrix as a heatmap
        sns.heatmap(confusion_matrix, annot=True, fmt='d',
                    cmap='Blues', cbar=False, ax=ax)

        # Set the title and axis labels
        ax.set_title(f'{title_prefix}Confusion Matrix for {classifier}')
        ax.set_xlabel('Predicted Label')
        ax.set_ylabel('True Label')

    # Adjust spacing between subplots
    plt.tight_layout()

    # Saves figure if required
    if save_fig:
        plt.savefig(filepath, dpi=300)

    # Show the plot
    plt.show()


def create_results_table(default_results, tuned_results):
    """
    Generate a results table comparing classifiers with and
    without hyperparameter tuning.

    Parameters:
        default_results (dict): A dictionary with classifier names
                                as keys and corresponding evaluation
                                metrics as values for the default models.
        tuned_results (dict): A dictionary with classifier names as keys,
                              where each value is another dictionary containing
                              the evaluation metrics for the tuned models.

    Returns:
        DataFrame: A DataFrame containing the performance metrics for each
                   classifier.
    """
    # Extract evaluation metrics and create DataFrame for default results
    default_metrics = {model: values['evaluation'] for model, values in
                       default_results.items()}
    default_df = pd.DataFrame(default_metrics).T
    default_df['Hyperparameter Tuning'] = 'No'

    # Extract evaluation metrics and create DataFrame for tuned results
    tuned_metrics = {model: values['evaluation'] for model, values in
                     tuned_results.items()}
    tuned_df = pd.DataFrame(tuned_metrics).T
    tuned_df['Hyperparameter Tuning'] = 'Yes'

    # Concatenate the two DataFrames
    results_df = pd.concat([default_df, tuned_df], axis=0).reset_index()
    results_df.rename(columns={'index': 'Classifier'}, inplace=True)

    # Reorder columns
    column_order = ['Classifier', 'Hyperparameter Tuning',
                    'specificity', 'recall', 'precision', 'accuracy',
                    'f1_score', 'roc_auc', 'cv_mean_recall', 'cv_std_recall']
    results_df = results_df[column_order]

    return results_df


def plot_shap(model, X_test, feature_names, plot_type='beeswarm'):
    """
    Plot SHAP values using a given model and test data.

    Parameters:
        model (object): A trained model.
        X_test (array): Test data.
        feature_names (list): List of feature names.
        plot_type (str): optional (default='beeswarm')
                         Type of SHAP plot to generate.

    Returns:
        None: This function will display a SHAP plot.
    """
    # If X_test is a numpy array, convert it to a DataFrame with feature names
    if isinstance(X_test, np.ndarray):
        X_test_df = pd.DataFrame(X_test, columns=feature_names)
    else:
        X_test_df = X_test

    # Create a SHAP explainer using the DataFrame
    explainer = shap.Explainer(model.predict, X_test_df)

    # Calculate SHAP values
    shap_values = explainer(X_test_df)

    # Create SHAP plot
    if plot_type == 'bar':
        shap.summary_plot(shap_values, X_test_df, plot_type=plot_type)
    elif plot_type == 'beeswarm':
        shap.plots.beeswarm(shap_values)
    else:
        raise ValueError("Invalid plot_type. Use 'bar' or 'beeswarm'.")

    plt.show()



# Import required libraries
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.decomposition import PCA
import pkg_resources

def load_data(filename):
    """
    Load data from a CSV file within the package into a pandas DataFrame.

    Parameters:
        filename (str): The filename of the CSV file to be read.

    Returns:
        DataFrame: DataFrame containing the data read from the CSV file.
    """
    filepath = pkg_resources.resource_filename('ML_medic_kit', f'data/{filename}')
    data = pd.read_csv(filepath)
    return data


def clean_dataset(data, outcome_name='diagnosis',
                  disease_outcome=['B', 'M'], binary=False,
                  to_drop=['id', 'Unnamed: 32'], encode=False):
    """
    Cleans the dataset by converting outcome column to binary if necessary,
    encoding any non-numeric predictor variables, renaming the outcome column,
    replacing spaces in column names, and dropping unnecessary columns.
    The default settings are suitable for cleaning the Wisconsin Breast Cancer
    dataset.

    Parameters:
        data (DataFrame): The dataset to be cleaned.
        outcome_name (str): optional (default="diagnosis")
            Name of disease outcome column.
        disease_outcome (list): optional (default=["B", "M"])
            Name of disease outcome.
        binary (bool): optional (default=False)
            If disease outcome names are binary.
        to_drop (list): optional (default=["id", "Unnamed: 32"])
            Names of columns to drop from dataset.
        encode: False (bool) or list, optional (default=False)
            False if no predictor variables need to be encoded,
            otherwise list of column names containing non-numeric
            values that need to be encoded.

    Returns:
        data (DataFrame): The cleaned dataset.
    """
    # Convert the outcome column to binary values
    if not binary:
        data[outcome_name] = data[outcome_name].map({disease_outcome[0]: 0,
                                                     disease_outcome[1]: 1})

    # Rename the outcome column
    data.rename(columns={outcome_name: 'outcome'}, inplace=True)

    # Drop unnecessary columns
    if len(to_drop) > 0:
        data.drop(columns=to_drop, inplace=True, errors='ignore')

    # Encodes other non-numeric columns using label encoder
    if encode is not False:
        for enc in encode:
            le = LabelEncoder()
            data[enc] = le.fit(data[enc]).transform(data[enc])

    # Replace spaces with underscores in column names
    data.columns = data.columns.str.replace(' ', '_')

    # Make column names all lower case
    data.columns = map(str.lower, data.columns)

    return data


def test_train_split_data(data, test_size):
    """
    The function separates the 'outcome' column from the rest of the data,
    then splits the dataset into training and test sets while stratifying based
    on the 'outcome' column to maintain the same proportion of classes.

    Parameters:
        data (DataFrame): The dataset to be split.
        test_size (int): The proportion to be test data

    Returns:
        tuple: Tuple containing the training and testing subsets
               (X_train, X_test, Y_train, Y_test).
    """

    # Separate features and the outcome variable
    Y = data['outcome']
    X = data.drop('outcome', axis=1)

    # Split dataset into training and test sets
    X_train, X_test, Y_train, Y_test = train_test_split(
        X, Y, test_size=test_size, stratify=Y, random_state=21
    )

    return X_train, X_test, Y_train, Y_test


def standard_scale_data(X_train, X_test):
    """
    Standard scales the data by removing the mean and scaling to unit variance.

    Parameters:
        X_train (DataFrame): Training data.
        X_test (DataFrame): Testing data.

    Returns:
        X_train_scaled (DataFrame): Standard scaled training data.
        X_test_scaled (DataFrame): Standard scaled testing data.
    """

    # Initialize StandardScaler
    scaler = StandardScaler()

    # Fit scaler on training data and transform it
    X_train_scaled = scaler.fit_transform(X_train)

    # Apply same transformation to test data
    X_test_scaled = scaler.transform(X_test)

    return X_train_scaled, X_test_scaled


def apply_pca(X_train, X_test, n_components):
    """
    Applies PCA dimensionality reduction to the training and test sets.

    Parameters:
        X_train (array): Training data.
        X_test (array): Test data.
        n_components (int): The number of principal components to keep.

    Returns:
        X_train_pca (array): PCA transformed training data.
        X_test_pca (array): PCA transformed test data.
    """
    pca = PCA(n_components=n_components)

    # Fit PCA on training data and transform test data
    X_train_pca = pca.fit_transform(X_train)
    X_test_pca = pca.transform(X_test)

    return X_train_pca, X_test_pca



import unittest
from sklearn.ensemble import RandomForestClassifier
from sklearn.datasets import make_classification
from ML_medic_kit.hyperparameter_tuning import tune_hyperparameters


class TestHyperparameterTuning(unittest.TestCase):
    """Tests for the hyperparameter_tuning module."""

    def setUp(self):
        # Create testing dataset
        self.X, self.y = make_classification(n_samples=100, n_features=5,
                                             random_state=42)

        # Create model
        self.model = RandomForestClassifier(random_state=42)

        #  Parameter grid for testing
        self.param_grid = {
            'n_estimators': [10, 50, 100],
            'max_depth': [None, 5, 10]
        }

    def test_tune_hyperparameters(self):
        # Perform hyperparameter tuning
        best_estimator, best_params, cv_results = tune_hyperparameters(
            self.model, self.param_grid, self.X, self.y)

        # Check if best_estimator is an instance of RandomForestClassifier
        self.assertIsInstance(best_estimator, RandomForestClassifier)

        # Check if best_params contains keys from param_grid
        self.assertTrue(all(key in self.param_grid for key in best_params))

        # Check if cv_results is a dictionary
        self.assertIsInstance(cv_results, dict)


if __name__ == '__main__':
    unittest.main()



# Import required libraries
import unittest
import pandas as pd
from ML_medic_kit import data_preprocessing
from sklearn.decomposition import PCA
import numpy as np
from sklearn.preprocessing import StandardScaler


class TestDataPreprocessing(unittest.TestCase):
    """Tests for the 'data_preprocessing' module."""

    def test_load_data(self):
        """Test the load_data function to ensure it returns a DataFrame."""

        FILEPATH = "ML_medic_kit/data/breast_cancer_data.csv"

        data = data_preprocessing.load_data(FILEPATH)

        # Check if the result is a DataFrame
        self.assertIsInstance(data, pd.DataFrame)

    def test_clean_dataset(self):
        """Test the clean_dataset function to ensure it cleans
        the data properly.
        """
        # Define the test data
        test_data = pd.DataFrame({
            'diagnosis': ['B', 'M', 'B', 'M'],
            'id': [1, 2, 3, 4],
            'Unnamed:_32': [5, 6, 7, 8],
            'feature': [9, 10, 11, 12]
        })
        cleaned_data = data_preprocessing.clean_dataset(test_data)

        # Check if 'id' column is dropped
        self.assertNotIn('id', cleaned_data.columns)

        # Check if 'Unnamed:_32' column is dropped
        self.assertNotIn('Unnamed:_32', cleaned_data.columns)

        # Check if 'diagnosis' is renamed to 'outcome'
        self.assertIn('outcome', cleaned_data.columns)

    def test_train_test_split_data(self):
        """Test the test_train_split_data function to ensure
        proper data splitting.
        """
        # Define the test data
        test_data = pd.DataFrame({
            'outcome': [0, 1, 0, 1],
            'feature1': [1, 2, 3, 4],
            'feature2': [5, 6, 7, 8]
        })

        X_train, X_test, Y_train, Y_test = \
            data_preprocessing.test_train_split_data(test_data, test_size=0.3)

        # Check if train size is 70% of 4
        self.assertEqual(len(X_train), 2)

        # Check if test size is 30% of 4
        self.assertEqual(len(X_test), 2)

    def test_standard_scale_data(self):
        """Test the standard scaling function on the feature data."""
        test_features = pd.DataFrame({
            'outcome': [0, 1, 0, 1],
            'feature1': [1, 2, 3, 4],
            'feature2': [10, 20, 30, 40]
        })

        X_train, X_test, Y_train, Y_test = \
            data_preprocessing.test_train_split_data(test_features,
                                                     test_size=0.5)

        # Initialize StandardScaler
        scaler = StandardScaler()

        # Fit scaler on training data and transform it
        X_train_scaled = scaler.fit_transform(X_train)

        # Apply transformation to test data
        X_test_scaled = scaler.transform(X_test)

        # Check that scaled data has zero mean and unit variance
        self.assertTrue(np.allclose(X_train_scaled.mean(axis=0), 0))
        self.assertTrue(np.allclose(X_train_scaled.std(axis=0), 1))

    def test_apply_pca(self):
        """Test the apply_pca function."""

        # Test data
        X_train = np.array([[1, 2], [3, 4], [5, 6]])
        X_test = np.array([[7, 8], [9, 10]])

        # Number of components to keep
        n_components = 2

        # Apply PCA
        X_train_pca, X_test_pca = data_preprocessing.apply_pca(X_train, X_test,
                                                               n_components)

        # Check if the dimensions are correct
        self.assertEqual(X_train_pca.shape[1], n_components)
        self.assertEqual(X_test_pca.shape[1], n_components)


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.datasets import make_classification
from unittest.mock import patch
from ML_medic_kit.visualisation import (
    plot_class_distribution,
    plot_explained_variance,
    correlation_heatmap,
    plot_roc,
    variables_based_outcome,
    histogram_based_outcome,
    plot_confusion_matrix,
    create_results_table,
    plot_shap
)


class TestVisualisation(unittest.TestCase):
    """Tests for visulaisation module"""

    def setUp(self):
        # Create date frame for testing
        data = {'Category': ['B', 'B', 'M', 'M', 'B', 'B', 'M', 'B']}
        self.df = pd.DataFrame(data)
        self.X_test = np.array([[1, 2, 3], [4, 5, 6]])
        self.feature_names = ['feature1', 'feature2', 'feature3']
        self.model = RandomForestClassifier(random_state=42)

        # Generate synthetic training data
        X_train, y_train = make_classification(n_samples=100,
                                               n_features=3,
                                               n_informative=2,
                                               n_redundant=0,
                                               n_repeated=0,
                                               random_state=42)

        # Fit the model with the synthetic training data
        self.model.fit(X_train, y_train)

    def test_plot_class_distribution(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            plot_class_distribution(self.df, 'Category',
                                    title='Test Distribution')
            show_patch.assert_called_once()

    def test_plot_explained_variance(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            plot_explained_variance(self.X_test)
            show_patch.assert_called_once()

    def test_create_results_table(self):
        default_results = {
            'Model1': {'evaluation': {'specificity': 0.8, 'recall': 0.7,
                                      'f1_score': 0.7, 'accuracy': 0.85,
                                      'precision': 0.8, 'roc_auc': 0.98,
                                      'cv_mean_recall': 0.7,
                                      'cv_std_recall': 0.02}},
            'Model2': {'evaluation': {'specificity': 0.8, 'recall': 0.7,
                                      'f1_score': 0.7, 'accuracy': 0.85,
                                      'precision': 0.8, 'roc_auc': 0.98,
                                      'cv_mean_recall': 0.7,
                                      'cv_std_recall': 0.02}},
        }
        tuned_results = {
            'Model1': {'evaluation': {'specificity': 0.8, 'recall': 0.7,
                                      'f1_score': 0.7, 'accuracy': 0.85,
                                      'precision': 0.8, 'roc_auc': 0.98,
                                      'cv_mean_recall': 0.7,
                                      'cv_std_recall': 0.02}},
            'Model2': {'evaluation': {'specificity': 0.8, 'recall': 0.7,
                                      'f1_score': 0.7, 'accuracy': 0.85,
                                      'precision': 0.8, 'roc_auc': 0.98,
                                      'cv_mean_recall': 0.7,
                                      'cv_std_recall': 0.02}},
        }
        results_df = create_results_table(default_results, tuned_results)

        # Check the number of rows in the data frame
        self.assertEqual(len(results_df), 4)

    def test_plot_confusion_matrix(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            results = {'Model1': {'confusion_matrix': [[1, 2], [3, 4]]},
                       'Model2': {'confusion_matrix': [[5, 6], [7, 8]]}}
            plot_confusion_matrix(results, title_prefix='Test')
            show_patch.assert_called_once()

    def test_correlation_heatmap(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            data = pd.DataFrame(np.random.randn(10, 3),
                                columns=['A', 'B', 'C'])
            correlation_heatmap(data)
            show_patch.assert_called_once()

    def test_plot_roc(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            X_test, y_test = make_classification(n_samples=100,
                                                 n_features=3,
                                                 n_informative=2,
                                                 n_redundant=0,
                                                 n_repeated=0,
                                                 random_state=42)
            models = {'RandomForest': self.model}
            plot_roc(models, X_test, y_test)
            show_patch.assert_called_once()

    def test_variables_based_outcome(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            data = pd.DataFrame({'var1': np.random.rand(10),
                                 'var2': np.random.rand(10),
                                 'outcome': np.random.randint(0, 2, 10)})
            variables_based_outcome(data, 'var1', 'var2')
            show_patch.assert_called_once()

    def test_histogram_based_outcome(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            data = pd.DataFrame({'feature1': np.random.rand(100),
                                 'feature2': np.random.rand(100),
                                 'outcome': np.random.randint(0, 2, 100)})
            histogram_based_outcome(data)
            show_patch.assert_called_once()

    def test_plot_shap(self):
        with patch("matplotlib.pyplot.show") as show_patch:
            X_test = np.random.rand(10, 3)
            feature_names = ['feature1', 'feature2', 'feature3']
            plot_shap(self.model, X_test, feature_names)
            # SHAP function generates two plots
            self.assertEqual(show_patch.call_count, 2)


if __name__ == '__main__':
    unittest.main()



# Import required libraries
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from ML_medic_kit.cross_validation import perform_cross_validation
from ML_medic_kit.model_evaluation import model_metrics
from ML_medic_kit.hyperparameter_tuning import tune_hyperparameters
import numpy as np


def create_classifier(classifier_type, X_train, y_train, RANDOM_SEED=42,
                      **params):
    """
    Create and fit a classifier based on the specified classifier_type.

    Parameters:
        classifier_type (str): Type of classifier to create.
        X_train (array): Training data.
        y_train (array): Target labels.
        **params: Additional hyperparameters to pass to the classifier.

    Returns:
        classifier: The trained classifier.
    Raises:
        ValueError: If an invalid classifier_type is provided.
    """
    # Create and fit a classifier based on the specified classifier_type
    if classifier_type == "LogisticRegression":
        model = LogisticRegression(random_state=RANDOM_SEED, **params)
    elif classifier_type == "RandomForest":
        model = RandomForestClassifier(random_state=RANDOM_SEED, **params)
    elif classifier_type == "MLPClassifier":
        model = MLPClassifier(random_state=RANDOM_SEED, **params)
    else:
        raise ValueError("Invalid classifier type")

    # Fit the model using the training data
    model.fit(X_train, y_train)

    # Return model
    return model


def train_test_multiple(classifiers, X_train, y_train, X_test, y_test,
                        param_grids=None, scoring='recall', cv=5,
                        tune=False, RANDOM_SEED=42):
    """
    Train and test multiple classifiers with or without hyperparameter tuning.

    Parameters:
        classifiers (list):  List of classifier names.
        X_train (array): Training features
        y_train (array): Training labels
        X_test (array): Test features
        param_grids (dict): Parameter grids for each classifier for
                            hyperparameter tuning.
        scoring: Scoring method to use for evaluation and cross-validation.
        cv (int): Number of cross-validation folds.
        tune: Whether to perform hyperparameter tuning.

    Returns:
        dict: A dictionary with evaluation results for each classifier.
        dict: A dictionary with the trained models
    """
    results = {}
    models = {}

    for classifier_type in classifiers:
        results[classifier_type] = {}

        # Create the model
        model = create_classifier(classifier_type, X_train, y_train,
                                  RANDOM_SEED=RANDOM_SEED)

        if tune and param_grids:
            # Perform hyperparameter tuning
            tuned_model, best_params, \
                grid_search_results = tune_hyperparameters(
                    model, param_grids[classifier_type], X_train,
                    y_train, scoring=scoring, cv=cv)
            results[classifier_type]['best_params'] = best_params
            cv_scores = grid_search_results['mean_test_score']
            model = tuned_model  # Use the tuned model for evaluation
        else:
            # Fit the model for default parameters
            model.fit(X_train, y_train)
            cv_scores = perform_cross_validation(model, X_train, y_train,
                                                 scoring=scoring, cv=cv)

        # Evaluate model performance
        metrics = model_metrics(model, X_test, y_test)
        metrics['cv_mean_recall'] = round(np.mean(cv_scores), 2)
        metrics['cv_std_recall'] = round(np.std(cv_scores), 2)

        results[classifier_type]['evaluation'] = metrics
        models[classifier_type] = model

        # Store evaluation metrics and confusion matrix
        results[classifier_type]['evaluation'] = metrics
        results[classifier_type]['confusion_matrix'] = \
            metrics['confusion_matrix']

    return results, models



# Imports required packages
from sklearn.metrics import confusion_matrix
from sklearn.metrics import roc_auc_score


def model_metrics(model, X_test, y_test):
    """
    Calculates model metrics for machine learning models.

    Parameters:
        model: Machine learning model.
        X_test (Dataframe): Test feature set.
        y_test (Dataframe): Actual disease outcome for test set.

    Returns:
        metric (dict):
            spec (float):
                Specificity - Out of all the people that do not have
                the disease, how many got negative results?
            recall (float):
                Recall (Sensitivity) - Out of all the people that
                have the disease, how many got positive test results?
            prec (float):
                Precision - Out of all the examples that predicted
                as positive, how many are really positive?
            acc (float):
                Accuracy - Fraction of predictions that the model
                got right.
            f1 (float):
                F1 score - Measure of the harmonic mean of
                precision and recall.
            roc_auc (float):
                Area under ROC curve.
    """
    # Make predictions on test data
    y_pred = model.predict(X_test)

    # Creates confusion matrix between predictions and actual results
    cm = confusion_matrix(y_test, y_pred)

    # Finds true positives/negatives and false positives/negatives
    TP = cm[0][0]
    FP = cm[0][1]
    TN = cm[1][1]
    FN = cm[1][0]

    # Calculates model scoring metrics
    spec = TN/(TN+FP)
    recall = TP/(TP+FN)
    prec = TP/(TP+FP)
    f1 = 2 * ((prec*recall) / (prec+recall))
    accuracy = (TP + TN) / (TP + FP + TN + FN)

    # Initialize roc_auc to None
    roc_auc = None

    # Check if the model has the predict_proba method for ROC-AUC calculation
    if hasattr(model, 'predict_proba'):
        y_pred_proba = model.predict_proba(X_test)[:, 1]
        roc_auc = roc_auc_score(y_test, y_pred_proba)

    # Puts scores in metrics dict
    metrics = {
        'specificity': round(spec, 2),
        'recall': round(recall, 2),
        'precision': round(prec, 2),
        'accuracy': round(accuracy, 2),
        'f1_score': round(f1, 2),
        'roc_auc': round(roc_auc, 2) if roc_auc is not None else None
    }

    # Add confusion matrix to metrics
    metrics['confusion_matrix'] = cm.tolist()

    return metrics


def find_best_model(default_results, tuned_results, metric):
    """
    Find the best model(s) based on a specified evaluation metric.

    Parameters:
        default_results (dict): Dictionary with default model results.
        tuned_results (dict): Dictionary with tuned model results.
        metric (str): The metric to compare.

    Returns:
        list: List of dictionaries containing information about
              the best model(s).

    Raises:
        ValueError: If the provided metric is not valid.
    """
    # Initialize variables
    best_models = []
    best_score = float('-inf')

    # Check the specified evaluation metric is in results
    # If not, raise a value error
    check1 = default_results[next(iter(default_results))]['evaluation']
    check2 = tuned_results[next(iter(tuned_results))]['evaluation']
    comment = f"The metric '{metric}' is not a valid evaluation metric."
    if metric not in check1:
        raise ValueError(comment)
    elif metric not in check2:
        raise ValueError(comment)

    # Check in default results
    for model, data in default_results.items():
        # Get the score for the specified metric
        score = data['evaluation'].get(metric, float('-inf'))
        # Check if this score is better or the same
        if score > best_score:
            best_models = [{"Best Model": model, "Best Score": score,
                            "Is Tuned": False, "Hyperparameters": "default"}]
            best_score = score
        elif score == best_score:
            best_models.append({"Best Model": model, "Best Score": score,
                                "Is Tuned": False,
                                "Hyperparameters": "default"})

    # Check in tuned results
    for model, data in tuned_results.items():
        # Get the score for the specified metric
        score = data['evaluation'].get(metric, float('-inf'))
        # Check if this score is better or the same
        if score > best_score:
            best_models = [{"Best Model": model, "Best Score": score,
                            "Is Tuned": True,
                            "Hyperparameters": data.get('best_params', None)}]
            best_score = score
        elif score == best_score:
            best_models.append({"Best Model": model, "Best Score": score,
                                "Is Tuned": True,
                                "Hyperparameters": data.get('best_params',
                                                            None)})

    return best_models



import unittest
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from ML_medic_kit import cross_validation


class TestCrossValidation(unittest.TestCase):
    """Tests for the cross_validation module."""

    def setUp(self):
        # Create dataset for testing
        self.X, self.y = make_classification(n_samples=100, n_features=4,
                                             n_classes=2, random_state=42)

    def test_perform_cross_validation(self):
        """Test the perform_cross_validation function."""
        # Create model for testing
        model = LogisticRegression()

        # Perform cross-validation
        scores = cross_validation.perform_cross_validation(model, self.X,
                                                           self.y,
                                                           scoring='recall',
                                                           cv=5)

        # Check if the scores array is not empty
        self.assertTrue(len(scores) > 0)

        # Check if the scores array has the correct length of 5 (cv = 5)
        self.assertEqual(len(scores), 5)


if __name__ == '__main__':
    unittest.main()



from sklearn.model_selection import cross_val_score
import numpy as np


def perform_cross_validation(model, X, y, scoring, cv=5):
    """
    Perform cross-validation on a given model and dataset.

    Parameters:
        model (object): The object to use to fit the data.
        X (array): The data to fit.
        y (array): The target variable to try to predict.
        scoring (str): A scorer  object
        cv (int): Determines the cross-validation splits.

    Returns:
        scores (array): Array of scores of the estimator for each
                        run of the cross validation.
    """
    scores = cross_val_score(model, X, y, scoring=scoring, cv=cv)

    return scores



import unittest
from sklearn.datasets import make_classification
from ML_medic_kit.model_selection import train_test_multiple, create_classifier
from sklearn.model_selection import train_test_split


class TestModelSelection(unittest.TestCase):
    """Tests for the model_selection module."""

    def setUp(self):
        # Create a dataset for testing
        self.X, self.y = make_classification(n_samples=30, n_features=5,
                                             random_state=42)

    def test_create_classifier(self):
        """Test creating classifiers with different types."""

        # Test that a valid classifier is created
        for classifier_type in ["LogisticRegression", "RandomForest",
                                "MLPClassifier"]:
            model = create_classifier(classifier_type, self.X, self.y)
            self.assertIsNotNone(model)

    def test_invalid_classifier(self):
        """Test creating a classifier with an invalid type."""
        # Test that creating an invalid classifier type raises a ValueError
        with self.assertRaises(ValueError):
            create_classifier("InvalidType", self.X, self.y)

    def test_train_test_multiple(self):
        """Test training and testing multiple classifiers."""

        # Split the dataset for training and testing
        X_train, X_test, y_train, y_test = train_test_split(self.X,
                                                            self.y,
                                                            test_size=0.3,
                                                            random_state=42)

        classifiers = ["LogisticRegression", "RandomForest", "MLPClassifier"]

        # default results
        results, _ = train_test_multiple(classifiers, X_train, y_train, X_test,
                                         y_test, cv=3)

        # Check that the function returns a dictionary
        self.assertIsInstance(results, dict)

        # Check that each classifier is in the results
        # and has an evaluation key
        for classifier in classifiers:
            self.assertIn(classifier, results)
            self.assertIn('evaluation', results[classifier])
            self.assertIn('cv_mean_recall',
                          results[classifier]['evaluation'])
            self.assertIn('cv_std_recall',
                          results[classifier]['evaluation'])

        # Simple param_grids for testing
        param_grids = {
            "LogisticRegression": {"C": [1]},
            "RandomForest": {"n_estimators": [10]},
            "MLPClassifier": {"hidden_layer_sizes": [(50,)]}
        }

        # tuned results
        tuned_results, _ = train_test_multiple(classifiers, X_train,
                                               y_train, X_test, y_test,
                                               param_grids=param_grids,
                                               scoring='recall', cv=3,
                                               tune=True)

        # Check that the function returns a dictionary
        self.assertIsInstance(tuned_results, dict)

        # Check that each classifier is in the results
        # and has an evaluation key
        for classifier in classifiers:
            self.assertIn(classifier, tuned_results)
            self.assertIn('evaluation', tuned_results[classifier])
            self.assertIn('best_params', tuned_results[classifier])
            self.assertIn('cv_mean_recall',
                          tuned_results[classifier]['evaluation'])
            self.assertIn('cv_std_recall',
                          tuned_results[classifier]['evaluation'])


if __name__ == '__main__':
    unittest.main()



import setuptools
from ML_medic_kit import __version__

# Read in the requirements.txt file
with open("requirements.txt") as f:
    requirements = []
    for library in f.read().splitlines():
        requirements.append(library)

# this is documentation...
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(

    name="ML_medic_kit",
    # set version number using value read in from __init__()
    version=__version__,
    author="Martha Dinsdale, Holly Manning",
    author_email="md661@exeter.ac.uk",
    license="The MIT License (MIT)",
    description="The Machine Learning Medic Kit is designed to enhance the capabilities of health data scientists tackling binary classification problems",
    # read in from readme.md and will appear on PyPi
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/marthadinsdale/hds_ca2",
    packages=setuptools.find_packages(),
    # if true look in MANIFEST.in for data files to include
    include_package_data=True,
    # these are for PyPi documentation
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.9',
    # pip requirements
    install_requires=requirements,
)



import unittest
from sklearn.ensemble import RandomForestClassifier
from sklearn.datasets import make_classification
from ML_medic_kit.model_evaluation import model_metrics, find_best_model


class TestModelEvaluation(unittest.TestCase):
    """Tests for the model_evaluation module."""

    # Define test data
    default_results = {
        'RandomForestClassifierDefault': {
            'evaluation': {
                'accuracy': 0.85,
                'f1_score': 0.9,
            }
        }
    }

    tuned_results = {
        'RandomForestClassifierTuned': {
            'evaluation': {
                'accuracy': 0.88,
                'f1_score': 0.92,
            },
            'best_params': {'n_estimators': 200, 'max_depth': 10}
        }
    }

    def setUp(self):
        # Create a dataset
        self.X, self.y = make_classification(n_samples=30, n_features=5,
                                             random_state=42)

        # Train a model
        self.model = RandomForestClassifier(n_estimators=100, random_state=42)
        self.model.fit(self.X, self.y)

    def test_model_metrics(self):
        # Calculate metrics using the model
        metrics = model_metrics(self.model, self.X, self.y)

        # Check if the metrics dictionary contains expected keys
        expected_keys = ['specificity', 'recall', 'precision',
                         'accuracy', 'f1_score', 'roc_auc', 'confusion_matrix']
        self.assertEqual(set(metrics.keys()), set(expected_keys))

    def test_find_best_model(self):
        # Find the best model based on accuracy
        best_models = find_best_model(self.default_results, self.tuned_results,
                                      'accuracy')

        # Check if the best_models list contains the expected model
        expected_best_models = [{'Best Model': 'RandomForestClassifierTuned',
                                 'Best Score': 0.88, 'Is Tuned': True,
                                 'Hyperparameters': {'n_estimators': 200,
                                                     'max_depth': 10}}]
        self.assertEqual(best_models, expected_best_models)


if __name__ == '__main__':
    unittest.main()



from sklearn.model_selection import GridSearchCV


def tune_hyperparameters(model, param_grid, X_train, y_train,
                         scoring=None, cv=5):
    """
    Perform hyperparameter tuning for a given model using GridSearchCV.

    Parameters:
        model (Object): The model for hyperparameter tuning
        param_grid (dict): Dictionary specifying the hyperparameter
                           grid to search.
        X_train (array): Training data.
        y_train (array): target labels
        scoring (str): A scorer object.
        cv (int): Determines the cross-validation splits.

    Returns:
        best_estimator (object): The best estimator found during
                                 the grid search.
        best_params (dict): The parameters of the best estimator
        cv_results (dict): Cross-validation results for the best estimator.

    """
    # If no scoring metric is provided, default to recall
    if scoring is None:
        scoring = 'recall'

    # Perform hyperparameter tuning using GridSearchCV
    grid_search = GridSearchCV(model, param_grid, scoring=scoring, cv=cv,
                               return_train_score=True)
    grid_search.fit(X_train, y_train)

    # Return the best estimator, best hyperparameters
    # and cross-validation results
    return (grid_search.best_estimator_, grid_search.best_params_,
            grid_search.cv_results_)
